"""Tests for any commands in readme.md."""

from authorityspoke.decisions import DecisionReading
import os
from dotenv import load_dotenv

load_dotenv()

from authorityspoke import CAPClient, Decision, LegisClient
from authorityspoke.io.fake_enactments import FakeClient

from authorityspoke.io.loaders import (
    load_decision,
    read_holdings_from_file,
    read_anchored_holdings_from_file,
)
from authorityspoke.io.loaders import read_holdings_from_file


class TestReadme:
    def test_posit_anchored_holdings(self, make_response):
        client = FakeClient(responses=make_response)

        oracle_dict = load_decision("oracle_h.json")
        lotus_dict = load_decision("lotus_h.json")
        oracle = Decision(**oracle_dict)
        lotus = Decision(**lotus_dict)

        oracle_ah = read_anchored_holdings_from_file(
            "holding_oracle.yaml", client=client
        )
        lotus_ah = read_anchored_holdings_from_file("holding_lotus.yaml", client=client)

        oracle_reading = DecisionReading(decision=oracle)
        lotus_reading = DecisionReading(decision=lotus)

        oracle_reading.posit(
            holdings=oracle_ah.holdings,
            named_anchors=oracle_ah.named_anchors,
            enactment_anchors=oracle_ah.enactment_anchors,
        )
        lotus_reading.posit(
            holdings=lotus_ah.holdings,
            named_anchors=lotus_ah.named_anchors,
            enactment_anchors=lotus_ah.enactment_anchors,
        )

        assert lotus_reading.contradicts(oracle_reading)

    def test_posit_holdings(self, make_response):
        client = FakeClient(responses=make_response)

        oracle_dict = load_decision("oracle_h.json")
        lotus_dict = load_decision("lotus_h.json")
        oracle = Decision(**oracle_dict)
        lotus = Decision(**lotus_dict)

        oracle_h = read_holdings_from_file("holding_oracle.yaml", client=client)
        lotus_h = read_holdings_from_file("holding_lotus.yaml", client=client)

        oracle_reading = DecisionReading(decision=oracle)
        lotus_reading = DecisionReading(decision=lotus)

        oracle_reading.posit(holdings=oracle_h)
        lotus_reading.posit(holdings=lotus_h)

        assert lotus_reading.contradicts(oracle_reading)

    def test_explain_contradiction(self, make_decision_with_holding):
        lotus = make_decision_with_holding["lotus"]
        oracle = make_decision_with_holding["oracle"]
        explanation = str(lotus.explain_contradiction(oracle))
        lotus_like = "<the Lotus menu command hierarchy> is like <the Java API>"
        java_like = "<the Java API> is like <the Lotus menu command hierarchy>"
        assert lotus_like in explanation or java_like in explanation



"""Functions for saving objects to file after they have been JSON serialized."""
import json
import pathlib

from typing import Dict, List, Optional, Union

from justopinion.decisions import Decision

from authorityspoke.decisions import Decision
from authorityspoke.io import filepaths


def case_to_file(
    case: Union[Decision, Decision],
    filename: Optional[str] = None,
    directory: Optional[pathlib.Path] = None,
    filepath: Optional[pathlib.Path] = None,
) -> None:
    r"""
    Save one case from an API response as a JSON file.

    :param results:
        A dict representing a case, in the format
        of the Caselaw Access Project API.

    :param filename:
        Filename (not including the directory) for the
        file where the downloaded opinion should be saved.

    :param directory:
        A :py:class:`~pathlib.Path` object specifying the directory where the
        downloaded opinion should be saved. If ``None`` is given, the current
        default is ``example_data/cases``\.

    :param filepath:
        Complete path to the location where the JSON file should be saved,
        including filename.
    """
    validated_filepath = filepaths.make_filepath(
        filename, directory, filepath, default_folder="cases"
    )
    with open(validated_filepath, "w") as fp:
        fp.write(case.json(indent=4))



from authorityspoke.holdings import HoldingGroup
from authorityspoke.opinions import OpinionReading
from copy import deepcopy
from datetime import date
import datetime
from pydantic import ValidationError
import pytest

from authorityspoke.decisions import Decision, DecisionReading, Opinion
from authorityspoke.io.loaders import read_anchored_holdings_from_file


class TestDecision:
    def test_decision_string(self, make_decision):
        decision = make_decision["cardenas"]
        assert str(decision) == "People v. Cardenas, 31 Cal. 3d 897 (1982-07-08)"

    def test_decision_no_opinions(self):
        decision = Decision(decision_date=date(2000, 2, 2))
        reading = DecisionReading(decision=decision)
        assert reading.majority is None
        assert not reading.implied_by(None)

    def test_decision_reading_with_opinion_instead_of_decision(self, make_opinion):
        with pytest.raises(ValidationError):
            DecisionReading(decision=make_opinion["lotus_majority"])

    def test_posit_holdings(self, make_decision, make_holding):
        """
        "Positing" Holdings causes the Holdings to be assigned to the Decision's
        majority Opinion.
        """

        watt = make_decision["watt"]
        reading = DecisionReading(decision=watt)
        reading.posit([make_holding["h1"], make_holding["h2"]])
        assert reading.majority.holdings[-1] == make_holding["h2"]

    def test_make_reading_with_holdings(self, make_decision, make_holding):
        watt = make_decision["watt"]
        holdings = HoldingGroup([make_holding["h1"], make_holding["h2"]])
        reading = DecisionReading(decision=watt)
        reading.posit(holdings)
        assert reading.majority.holdings[-1] == make_holding["h2"]

    def test_make_reading_with_anchored_holdings(
        self, make_decision, make_anchored_holding
    ):
        watt = make_decision["watt"]
        holding = make_anchored_holding["lotus"]
        reading = DecisionReading(decision=watt)
        reading.posit(holding)
        assert (
            str(reading.majority.holdings[0].inputs[0]).lower()
            == "the fact that <the lotus menu command hierarchy> was copyrightable"
        )
        assert reading.majority.opinion_author.startswith("HAMLEY")
        assert reading.decision.opinions[0].author.startswith("HAMLEY")

    def test_need_reading_to_posit_holding(self, make_holding):
        decision = Decision(decision_date=(datetime.date(1900, 1, 1)))
        with pytest.raises(AttributeError):
            decision.posit(make_holding["h1"])

    def test_decision_reading_has_opinion_readings(self, make_decision, make_holding):
        watt = make_decision["watt"]
        reading = DecisionReading(decision=watt)
        reading.posit(make_holding["h1"])
        hamley = "hamley, circuit judge"
        assert reading.opinion_readings[0].opinion_author.lower() == hamley
        assert reading.opinion_readings[0].holdings[0].means(make_holding["h1"])

    def test_decision_posits_holding(self, fake_usc_client, make_decision):
        lotus_analysis = read_anchored_holdings_from_file(
            "holding_lotus.yaml", client=fake_usc_client
        )
        lotus_reading = DecisionReading(decision=make_decision["lotus"])
        lotus_reading.posit(lotus_analysis)
        assert len(lotus_reading.majority.holdings) == len(lotus_analysis.holdings)
        assert str(lotus_reading).startswith("Reading for Lotus Development Corp.")

    def test_decision_with_opinion_reading_posits_holding(self, fake_usc_client):
        lotus_analysis = read_anchored_holdings_from_file(
            "holding_lotus.yaml", client=fake_usc_client
        )
        decision_reading = DecisionReading(
            decision=Decision(decision_date=date(2000, 2, 2)),
            opinion_readings=[OpinionReading(opinion_type="plurality")],
        )
        decision_reading.posit(lotus_analysis)
        assert len(decision_reading.holdings) == len(lotus_analysis.holdings)

    def test_error_decision_with_no_majority_posits_holding(self, fake_usc_client):
        lotus_analysis = read_anchored_holdings_from_file(
            "holding_lotus.yaml", client=fake_usc_client
        )
        reading1 = OpinionReading(opinion_type="plurality")
        reading2 = OpinionReading(opinion_type="concurring")
        decision_reading = DecisionReading(
            decision=Decision(decision_date=date(2000, 2, 2)),
            opinion_readings=[reading1, reading2],
        )
        with pytest.raises(AttributeError):
            decision_reading.posit(lotus_analysis)

    def test_add_opinion_to_decision(
        self, make_decision, make_opinion, make_opinion_with_holding
    ):
        reading = DecisionReading(decision=Decision(decision_date=date(2000, 2, 2)))
        assert len(reading.opinions) == 0
        reading.add_opinion(make_opinion["oracle_majority"])
        assert len(reading.opinions) == 1

    def test_add_opinion_with_holdings(
        self, make_decision, make_opinion, make_opinion_with_holding
    ):
        reading = DecisionReading(decision=make_decision["oracle"])
        assert len(reading.opinion_readings) == 0
        assert len(reading.holdings) == 0
        reading.add_opinion_reading(make_opinion_with_holding["oracle_majority"])
        assert len(reading.opinion_readings) == 1
        assert len(reading.holdings) == 20


class TestImplication:
    def test_implication_of_decision_with_one_of_same_holdings(
        self, make_decision, make_decision_with_holding
    ):
        decision = make_decision["oracle"]
        oracle = DecisionReading(decision=decision)
        oracle_with_holdings = make_decision_with_holding["oracle"]
        oracle.posit(oracle_with_holdings.holdings[0])
        assert len(oracle.holdings) == 1
        assert len(oracle_with_holdings.holdings) > 10
        assert oracle_with_holdings >= oracle

    def test_decision_implied_by_holding(self, make_decision, make_holding):
        decision = make_decision["watt"]
        reading = DecisionReading(decision=decision)
        holding = make_holding["h1"]
        reading.posit(holding)
        assert reading.implied_by(holding)

    def test_decision_explain_implication(
        self, make_decision_with_holding, make_holding
    ):
        decision = make_decision_with_holding["watt"]
        holding = decision.holdings[0]
        explanation = decision.explain_implication(holding)
        assert explanation

    def test_decision_no_explanation_implication(self, make_decision_with_holding):
        watt = make_decision_with_holding["watt"]
        oracle = make_decision_with_holding["oracle"]
        explanation = watt.explain_implication(oracle)
        assert explanation is None

    def test_typeerror_to_compare_with_factor(
        self, make_decision_with_holding, make_factor
    ):
        watt = make_decision_with_holding["watt"]
        factor = make_factor["f_no_shooting"]
        with pytest.raises(TypeError):
            watt.explain_implication(factor)

    def test_decision_implies_decision_without_holdings(
        self, make_decision_with_holding, make_decision
    ):
        oracle = make_decision_with_holding["oracle"]
        blank = make_decision["lotus"]
        blank_reading = DecisionReading(decision=blank)
        explanation = oracle.explain_implication(blank_reading)
        assert not explanation.reasons

    def test_no_holdings_of_blank_decision(self):
        blank = Decision(decision_date=datetime.date(2000, 1, 2))
        reading = DecisionReading(decision=blank)
        assert len(reading.holdings) == 0

    # @pytest.mark.skip(reason="slow")
    def test_decision_implies_its_opinion(self, make_decision_with_holding):
        cardenas = make_decision_with_holding["cardenas"]
        assert cardenas >= cardenas.majority
        assert cardenas > cardenas.majority

    # @pytest.mark.skip(reason="slow")
    def test_opinion_implies_its_decision(self, make_decision_with_holding):
        cardenas = make_decision_with_holding["cardenas"]
        assert cardenas.implied_by(cardenas.majority)

    def test_decision_implies_holding_and_rule(self, make_decision_with_holding):
        """Adding a new input makes the new Holding less specific than the original one."""

        oracle = make_decision_with_holding["oracle"]
        holding = deepcopy(oracle.holdings[1])
        new_inputs = holding.inputs + [oracle.holdings[0].inputs[0]]
        holding.set_inputs(new_inputs)
        assert oracle.implies(holding)
        assert oracle.implies(holding.rule)

    def test_decision_does_not_imply_procedure(
        self, make_procedure, make_decision_with_holding
    ):
        oracle = make_decision_with_holding["oracle"]
        assert not oracle.implies(make_procedure["c1"])

    def test_opinion_implies_decision(self, make_decision_with_holding):
        watt = make_decision_with_holding["watt"]
        assert watt.opinion_readings[0].implies(watt)

    def test_rule_implied_by_decision(self, make_decision_with_holding):
        oracle = make_decision_with_holding["oracle"]
        explanation = oracle.explain_implication(oracle.holdings[0].rule)
        assert len(explanation.reasons) == 3

    def test_decision_not_implied_by_rule(self, make_decision_with_holding):
        oracle = make_decision_with_holding["oracle"]
        explanation = oracle.explain_implied_by(oracle.holdings[0].rule)
        assert explanation is None


class TestContradiction:
    def test_oracle_contradicts_lotus(self, make_decision_with_holding):
        """
        This is the same example from the "Introduction" notebook.
        """
        oracle = make_decision_with_holding["oracle"]
        lotus = make_decision_with_holding["lotus"]
        assert oracle.contradicts(lotus)
        assert lotus.contradicts(oracle)

    def test_no_contradiction_explanations(self, make_decision_with_holding):
        oracle = make_decision_with_holding["oracle"]
        feist = make_decision_with_holding["feist"]
        explanation = oracle.explain_contradiction(feist)
        assert explanation is None

    def test_cannot_check_contradiction_of_str(self, make_decision_with_holding):
        oracle = make_decision_with_holding["oracle"]
        with pytest.raises(TypeError):
            oracle.explain_contradiction("feist")

    def test_no_contradiction_with_plurality(self, make_decision_with_holding):
        oracle = make_decision_with_holding["oracle"]
        other = Decision(
            decision_date=datetime.date(2020, 1, 1),
            opinions=[Opinion(position="plurality")],
        )
        reading = DecisionReading(decision=other)
        assert not oracle.contradicts(reading)

    def test_no_contradiction_of_majority_without_holdings(
        self, make_decision_with_holding
    ):
        oracle = make_decision_with_holding["oracle"]
        other = Decision(decision_date=datetime.date(2020, 1, 1))
        other.opinions.append(Opinion(position="majority"))
        reading = DecisionReading(decision=other)
        assert not oracle.contradicts(reading)

    def test_decision_contradicts_holding(self, make_decision_with_holding):
        oracle = make_decision_with_holding["oracle"]
        holding = oracle.holdings[0].negated()
        assert oracle.contradicts(holding)



r"""
Interpretations (or "readings") of judicial Decisions.

Each :class:`.justopinion.decisions.Decision` may include
multiple :class:`.justopinion.decisions.Opinion`\s. A
:class:`.authorityspoke.decision.DecisionReading` may link
to multiple :class:`.authorityspoke.opinion.OpinionReading`\s.
"""

from __future__ import annotations

import operator
from typing import Dict, Iterator, List
from typing import Optional, Sequence, Tuple, Union

from anchorpoint.textselectors import TextQuoteSelector, TextPositionSelector
from justopinion.decisions import Decision, CaseBody, CaseData, Opinion, CAPCitation

from nettlesome.terms import Comparable, ContextRegister, Explanation
from nettlesome.factors import Factor
from pydantic import BaseModel

from authorityspoke.holdings import Holding, HoldingGroup
from authorityspoke.opinions import (
    OpinionReading,
    TermWithAnchors,
    AnchoredHoldings,
    HoldingWithAnchors,
    EnactmentWithAnchors,
)
from authorityspoke.rules import Rule


RawCAPCitation = Dict[str, str]
RawOpinion = Dict[str, str]
RawDecision = Dict[str, Union[str, int, Sequence[RawOpinion], Sequence[RawCAPCitation]]]


class DecisionReading(BaseModel, Comparable):
    """An interpretation of what Holdings are supported by the Opinions of a Decision."""

    decision: Decision
    opinion_readings: List[OpinionReading] = []

    def __str__(self):
        citation = self.decision.citations[0].cite if self.decision.citations else ""
        name = self.decision.name_abbreviation or self.decision.name
        return f"Reading for {name}, {citation} ({self.decision.decision_date})"

    @property
    def majority(self) -> Optional[OpinionReading]:
        """Return the majority OpinionReading, or None if it doesn't exist."""
        for reading in self.opinion_readings:
            if reading.opinion_type == "majority":
                return reading
        return None

    @property
    def opinions(self) -> List[Opinion]:
        """Get the Opinions for this DecisionReading's Decision."""
        return self.decision.opinions

    def find_matching_opinion(
        self,
        opinion_type: str = "",
        opinion_author: str = "",
    ) -> Optional[Opinion]:
        """Find an Opinion described by the given attributes."""
        return self.decision.find_matching_opinion(
            opinion_type=opinion_type, opinion_author=opinion_author
        )

    def find_opinion_matching_reading(
        self,
        opinion_reading: OpinionReading,
    ) -> Optional[Opinion]:
        """Find the Opinion corresponding to an OpinionReading."""
        return self.find_matching_opinion(
            opinion_type=opinion_reading.opinion_type,
            opinion_author=opinion_reading.opinion_author,
        )

    def select_text(
        self,
        selector: Union[
            bool,
            str,
            TextPositionSelector,
            TextQuoteSelector,
            Sequence[
                Union[str, Tuple[int, int], TextQuoteSelector, TextPositionSelector]
            ],
        ],
        opinion_type: str = "",
        opinion_author: str = "",
    ) -> Optional[str]:
        r"""
        Get text using a :class:`.TextQuoteSelector`.

        :param selector:
            a selector referencing a text passage in the :class:`Opinion`.

        :returns:
            the text referenced by the selector, or ``None`` if the text
            can't be found.
        """
        opinion = self.find_matching_opinion(opinion_type, opinion_author)
        return opinion.select_text(selector) if opinion else None

    def add_opinion_reading(self, opinion_reading: OpinionReading) -> None:
        """Add an OpinionReading for an existing Opinion of the Decision."""
        matching_opinion = self.find_opinion_matching_reading(opinion_reading)
        if matching_opinion:
            opinion_reading.opinion_type = (
                matching_opinion.type or opinion_reading.opinion_type
            )
            opinion_reading.opinion_author = (
                matching_opinion.author or opinion_reading.opinion_author
            )
        self.opinion_readings.append(opinion_reading)

    def get_majority(self) -> Optional[OpinionReading]:
        """Return the majority OpinionReading, creating it if needed."""
        majority = self.majority
        if majority:
            return majority
        for opinion in self.decision.opinions:
            if opinion.type == "majority":
                new_reading = OpinionReading(
                    opinion_type=opinion.type, opinion_author=opinion.author
                )
                self.opinion_readings.append(new_reading)
                return new_reading
        return None

    @property
    def holdings(self) -> HoldingGroup:
        """Get the holdings of this Decision's majority Opinion."""
        if self.majority is not None:
            return HoldingGroup(self.majority.holdings)
        elif len(self.opinion_readings) == 1:
            return HoldingGroup(self.opinion_readings[0].holdings)
        return HoldingGroup()

    def add_opinion(self, opinion: Opinion) -> None:
        """Link an Opinion document to this Decision."""
        if not self.decision.casebody:
            self.decision.casebody = CaseBody(data=CaseData())
        self.decision.casebody.data.opinions.append(opinion)

    def contradicts(self, other):
        """Check if a holding attributed to this decision contradicts a holding attributed in "other"."""
        if isinstance(other, DecisionReading):
            if self.majority and other.majority:
                return self.majority.contradicts(other.majority)
            return False
        return self.majority.contradicts(other)

    def explain_contradiction(
        self, other: Union[OpinionReading, Holding, Rule]
    ) -> Optional[Explanation]:
        """Get the first generated explanation of how a Holding of self contradicts a Holding of other."""
        explanations = self.explanations_contradiction(other)
        try:
            explanation = next(explanations)
        except StopIteration:
            return None
        return explanation

    def explanations_contradiction(
        self,
        other: Union[DecisionReading, Opinion, Holding, Rule],
    ) -> Iterator[Explanation]:
        """Generate explanations of how a Holding of self contradicts a Holding of other."""
        if isinstance(other, DecisionReading):
            if self.majority and other.majority:
                yield from self.majority.explanations_contradiction(other.majority)
        elif isinstance(other, (Rule, Holding, OpinionReading)):
            if self.majority:
                yield from self.majority.explanations_contradiction(other)
        else:
            raise TypeError(
                f"'Contradicts' test not implemented for types "
                f"{self.__class__} and {other.__class__}."
            )

    def explain_implication(
        self,
        other: Union[Opinion, Holding, Rule],
    ) -> Optional[Explanation]:
        """Get the first generated explanation of how a Holding of self implies a Holding of other."""
        explanations = self.explanations_implication(other)
        try:
            explanation = next(explanations)
        except StopIteration:
            return None
        return explanation

    def explanations_implication(
        self, other: Union[DecisionReading, Decision, Opinion, Holding, Rule]
    ) -> Iterator[Explanation]:
        """Generate explanation of how self's Holdings can imply other."""
        if isinstance(other, DecisionReading):
            self_majority = self.get_majority()
            if self_majority and other.get_majority():
                yield from self_majority.explanations_implication(other.majority)
        elif isinstance(other, (Rule, Holding, OpinionReading)):
            self_majority = self.get_majority()
            if self_majority:
                yield from self_majority.explanations_implication(other)
        else:
            raise TypeError(
                f"'Implication' test not implemented for types "
                f"{self.__class__} and {other.__class__}."
            )

    def posit(
        self,
        holdings: Union[
            AnchoredHoldings,
            Holding,
            Rule,
            HoldingWithAnchors,
            List[Union[HoldingWithAnchors, Holding, Rule]],
        ],
        holding_anchors: Optional[List[HoldingWithAnchors]] = None,
        named_anchors: Optional[List[TermWithAnchors]] = None,
        enactment_anchors: Optional[List[EnactmentWithAnchors]] = None,
        context: Optional[Sequence[Factor]] = None,
    ) -> None:
        """Add one or more Holdings to the majority Opinion of this Decision."""
        majority = self.get_majority()
        if majority is not None:
            reading_to_use = majority
        elif len(self.opinion_readings) == 1:
            reading_to_use = self.opinion_readings[0]
        elif not self.opinion_readings:
            self.opinion_readings = [OpinionReading()]
            reading_to_use = self.opinion_readings[0]
        else:
            raise AttributeError(
                "Cannot determine how to posit the Holding because this DecisionReading "
                "has multiple OpinionReadings, none of which is linked to a majority "
                "Opinion. Try using the .posit() method of one of the OpinionReadings "
                "on this object's `opinion_readings` attribute."
            )
        reading_to_use.posit(
            holdings=holdings,
            holding_anchors=holding_anchors,
            named_anchors=named_anchors,
            enactment_anchors=enactment_anchors,
            context=context,
        )

    def __ge__(self, other) -> bool:
        return self.implies(other)

    def __gt__(self, other) -> bool:
        return self.implies(other) and self != other

    def implied_by_holding(
        self, other: Holding, context: Optional[ContextRegister] = None
    ) -> Iterator[Explanation]:
        """Check if a Holding implies all the Holdings of self."""
        if all(
            other.implies(self_holding, context=context)
            for self_holding in self.holdings
        ):
            yield Explanation(
                reasons=[(other, self_holding) for self_holding in self.holdings],
                operation=operator.ge,
            )

    def explanations_implied_by(
        self, other: Comparable, context: Optional[ContextRegister] = None
    ) -> Iterator[Explanation]:
        """Generate explanation of how other can imply all the Holdings of self."""
        context = context or ContextRegister()
        if isinstance(other, OpinionReading):
            other = other.holdings
        if isinstance(other, HoldingGroup):
            yield from other.explanations_implication(
                self.holdings, context=context.reversed()
            )
        if isinstance(other, Rule):
            other = Holding(rule=other)
        if isinstance(other, Holding):
            yield from self.implied_by_holding(other, context)

    def implied_by(
        self, other: Optional[Comparable], context: Optional[ContextRegister] = None
    ) -> bool:
        """Check if other implies all the Holdings of self."""
        if other is None:
            return False
        return any(self.explanations_implied_by(other, context))

    def implies_holding(
        self, other: Holding, context: Optional[ContextRegister] = None
    ) -> bool:
        """Check if a Holding of self implies a Holding made from other."""
        return any(
            self_holding.implies(other, context=context)
            for self_holding in self.holdings
        )

    def implies_rule(
        self, other: Rule, context: Optional[ContextRegister] = None
    ) -> bool:
        """Check if a Holding of self implies a Holding made from other Rule."""
        return self.implies_holding(Holding(rule=other), context=context)

    def implies(self, other, context: Optional[ContextRegister] = None) -> bool:
        """Check if the Holdings of self imply other."""
        if isinstance(other, (DecisionReading, OpinionReading)):
            return self.holdings.implies(other.holdings)
        elif isinstance(other, Holding):
            return self.implies_holding(other, context=context)
        elif isinstance(other, Rule):
            return self.implies_rule(other, context=context)
        return False



import pytest

from anchorpoint.textselectors import (
    TextPositionSet,
    TextQuoteSelector,
    TextSelectionError,
)

from nettlesome.entities import Entity

from authorityspoke.facts import Fact, Predicate
from authorityspoke.holdings import Holding, HoldingGroup
from authorityspoke.procedures import Procedure
from authorityspoke.rules import Rule
from authorityspoke.io import loaders, readers
from authorityspoke.io.fake_enactments import FakeClient
from authorityspoke.opinions import (
    AnchoredHoldings,
    Opinion,
    FactorIndex,
    OpinionReading,
    TermWithAnchors,
)


class TestOpinions:
    def test_opinion_features(self, make_decision):
        assert make_decision["watt"].court.slug == "9th-cir"
        assert make_decision["watt"].citations[0].cite == "388 F.2d 853"

    def test_repr(self, make_opinion):
        assert "HAMLEY, Circuit Judge" in repr(make_opinion["watt_majority"])

    def test_repr_excludes_text(self, make_opinion):
        """
        reprs would be way too long if they could contain the
        full opinion text.
        """
        assert "text" not in repr(make_opinion["watt_majority"])

    def test_opinion_author(self, make_opinion):
        assert make_opinion["watt_majority"].author == "HAMLEY, Circuit Judge"
        assert make_opinion["brad_majority"].author == "BURKE, J."
        assert (
            make_opinion["brad_concurring-in-part-and-dissenting-in-part"].author
            == "TOBRINER, J."
        )

    def test_opinion_holding_list(self, make_opinion_with_holding, real_holding):
        watt = make_opinion_with_holding["watt_majority"]
        h3_specific = real_holding["h3"]
        watt.posit(h3_specific)
        assert h3_specific in watt.holdings

    def test_opinion_entity_list(
        self, make_opinion, real_holding, make_entity, make_evidence
    ):
        watt = make_opinion["watt_majority"]
        h = real_holding
        e = make_entity
        reading = OpinionReading(opinion_type=watt.type, opinion_author=watt.author)
        reading.posit(h["h1"], context=(e["motel"], e["watt"]))
        reading.posit(h["h2"], context=(e["trees"], e["motel"]))
        reading.posit(
            h["h3"],
            context=(
                make_evidence["generic"],
                e["motel"],
                e["watt"],
                e["trees"],
                e["tree_search"],
            ),
        )
        reading.posit(
            h["h4"], context=(e["trees"], e["tree_search"], e["motel"], e["watt"])
        )
        assert make_entity["watt"] in reading.generic_terms()

    def test_opinion_date(self, make_decision):
        """
        This no longer tests whether two Opinions from the same Decision
        have the same date, because the date field has been moved to the
        Decision class.
        """
        assert make_decision["watt"].decision_date < make_decision["brad"].decision_date

    def test_opinion_string(self, make_opinion):
        opinion = make_opinion["cardenas_majority"]
        assert str(opinion).lower() == "majority opinion by bird, c. j."


class TestOpinionText:
    def test_opinion_text(self, make_opinion):
        assert (
            "Feist responded that such efforts were economically "
            + "impractical and, in any event, unnecessary"
        ) in make_opinion["feist_majority"].text

    def test_opinion_text_anchor(self, make_opinion_with_holding):
        feist = make_opinion_with_holding["feist_majority"]
        assert any(
            "ideas" in factor.anchors.quotes[0].exact
            for factor in feist.anchored_factors
        )

    def test_select_opinion_text_for_factor(self, make_opinion, make_anchored_holding):
        opinion = make_opinion["oracle_majority"]
        oracle = make_anchored_holding["oracle"]
        anchor = oracle.get_term_anchors(
            key="the fact that <the Java API> was copyrightable"
        )
        selected = opinion.select_text(selector=anchor)
        assert str(selected).startswith("…copyright protection.")

    def test_select_opinion_text_for_enactment(self, make_decision_with_holding):
        oracle = make_decision_with_holding["oracle"]
        enactment_name = str(oracle.holdings[0].enactments[0])
        anchor = oracle.majority.get_enactment_anchors(enactment_name)
        selected = oracle.select_text(selector=anchor, opinion_type="majority")
        assert str(selected) == "…17 U.S.C. § 102(a)…"

    def test_select_opinion_text_for_holding(self, make_opinion, make_anchored_holding):
        opinion = make_opinion["oracle_majority"]
        anchor = make_anchored_holding["oracle"].holdings[0].anchors
        selected = opinion.select_text(selector=anchor)
        assert str(selected) == "…must be “original” to qualify…"

    def test_invalid_text_selector(self, make_decision_with_holding):
        oracle = make_decision_with_holding["oracle"]
        anchor = TextQuoteSelector(exact="text not in opinion")
        with pytest.raises(TextSelectionError):
            _ = oracle.decision.majority.select_text(selector=anchor)


class TestOpinionHoldings:
    def test_positing_non_rule_error(self, make_opinion_with_holding, make_procedure):
        with pytest.raises(TypeError):
            make_opinion_with_holding["watt_majority"].posit(make_procedure["c1"])

    def test_error_posit_with_no_rule_source(self, make_opinion_with_holding):
        with pytest.raises(TypeError):
            make_opinion_with_holding["watt_majority"].posit()

    def test_posit_rule(self, make_opinion_with_holding, make_rule, make_holding):
        """
        "Positing" a Rule causes the Rule to be converted to a Holding first.
        So the Opinion implies the corresponding Holding.
        """

        watt = make_opinion_with_holding["watt_majority"]
        watt.clear_holdings()
        watt.posit(make_rule["h1"])
        assert watt.implies(make_holding["h1"])

    def test_new_context_wrong_number_of_changes(
        self, make_opinion_with_holding, make_holding
    ):
        brad = make_opinion_with_holding["brad_majority"]
        with pytest.raises(ValueError):
            brad.posit(
                make_holding["h1"], context=[Entity(name="House on Haunted Hill")]
            )

    def test_new_context_naming_nonexistent_factor(
        self, make_opinion_with_holding, make_holding
    ):
        """
        The context here (a Factor outside an iterable) only changes the first
        generic factor of the Rule being posited, which may not be what the user
        expects.
        """
        brad = make_opinion_with_holding["brad_majority"]
        brad.clear_holdings()
        with pytest.raises(ValueError):
            brad.posit(
                make_holding["h1"],
                context=(Entity(name="House on Haunted Hill"), "nonexistent factor"),
            )

    def test_new_context_creates_equal_rule(
        self, make_opinion_with_holding, make_response
    ):
        watt = make_opinion_with_holding["watt_majority"]
        brad = make_opinion_with_holding["brad_majority"]
        client = FakeClient(responses=make_response)

        watt.clear_holdings()
        watt_raw = loaders.load_holdings("holding_watt.yaml")
        watt.posit(readers.read_holdings(watt_raw, client=client))

        brad.clear_holdings()
        brad_raw = loaders.load_holdings("holding_brad.yaml")
        brad.posit(readers.read_holdings(brad_raw, client=client))

        context_pairs = {
            "proof of Bradley's guilt": "proof of Wattenburg's guilt",
            "Bradley": "Wattenburg",
            "officers' search of the yard": "officers' search of the stockpile",
            "Bradley's marijuana patch": "the stockpile of trees",
        }
        watt.posit(brad.holdings[0], context_pairs)
        assert watt.holdings[-1].means(brad.holdings[0])

    def test_getting_factors_from_opinion(
        self, make_opinion_with_holding, make_response
    ):
        client = FakeClient(responses=make_response)

        watt = make_opinion_with_holding["watt_majority"]
        watt.clear_holdings()
        watt_raw = loaders.load_holdings("holding_watt.yaml")
        holdings_to_posit = readers.read_holdings(watt_raw, client=client)
        watt.posit(holdings_to_posit)
        factors = watt.factors_by_name()
        assert "proof of Wattenburg's guilt" in factors.keys()

    def test_new_context_inferring_factors_to_change(
        self, make_opinion_with_holding, make_response
    ):
        """
        This changes watt's holdings; may break tests below.
        """
        watt = make_opinion_with_holding["watt_majority"]
        brad = make_opinion_with_holding["brad_majority"]

        client = FakeClient(responses=make_response)

        watt.clear_holdings()
        watt_raw = loaders.load_holdings("holding_watt.yaml")
        watt.posit(readers.read_holdings(watt_raw, client=client))

        brad.clear_holdings()
        brad_raw = loaders.load_holdings("holding_brad.yaml")
        brad.posit(readers.read_holdings(brad_raw, client=client))

        context_items = [
            "proof of Wattenburg's guilt",
            "<Wattenburg>",
            "<officers' search of the stockpile>",
            "<Hideaway Lodge>",
            "<the stockpile of trees>",
        ]
        watt.posit(brad.holdings[0], context=context_items)
        assert watt.holdings[-1].means(brad.holdings[0])

    def test_getting_factors_from_new_holding(self, make_opinion_with_holding):
        watt = make_opinion_with_holding["watt_majority"]
        watt.clear_holdings()
        elephants = Fact(
            predicate="$animal was an elephant", terms=Entity(name="the elephant")
        )
        mouseholes = Fact(
            predicate=Predicate(content="$animal hides in mouseholes", truth=False),
            terms=Entity(name="the elephant"),
        )
        procedure = Procedure(inputs=elephants, outputs=mouseholes)
        rule = Rule(procedure=procedure)
        holding = Holding(rule=rule)
        watt.posit(holding)
        factors = watt.factors_by_name()
        factor = factors["the fact that <the elephant> was an elephant"]
        assert factor.terms[0].name == "the elephant"


class TestOpinionFactors:
    def test_only_one_factor_with_same_content(self, make_opinion_with_holding):
        """
        Tests that a particular Factor appears only once, and that all
        three of the text anchors for that Factor appear in the value
        for the Factor in Opinion.factors.
        """

        oracle = make_opinion_with_holding["oracle_majority"]
        scenes_a_faire = [
            factor
            for factor in oracle.factors()
            if isinstance(factor, Fact)
            and factor.short_string
            == "the fact that <the Java API> was a scene a faire"
        ]
        assert len(scenes_a_faire) == 1  # 1 Factor

    def test_insert_duplicate_anchor_in_factor_index(self):
        api = Entity(name="the Java API", generic=True, plural=False)
        quote_selector = TextQuoteSelector(
            exact="it possesses at least some minimal degree of creativity."
        )
        anchors = TextPositionSet(quotes=quote_selector)
        fact = Fact(
            predicate=Predicate(
                content="$product possessed at least some minimal degree of creativity"
            ),
            terms=[api],
        )
        term = TermWithAnchors(term=fact, anchors=anchors)
        index = AnchoredHoldings(holdings=[], named_anchors=[term])

        index.add_term(term=fact, anchors=anchors)
        assert len(index.named_anchors) == 1
        assert index.named_anchors[0].anchors.quotes == [quote_selector]

    def test_get_factor_from_opinion(self, make_opinion_with_holding):
        oracle = make_opinion_with_holding["oracle_majority"]
        company = oracle.get_factor_by_name("the Java API")
        assert isinstance(company, Entity)

    def test_factors_by_name(self, make_opinion_with_holding):
        oracle = make_opinion_with_holding["oracle_majority"]
        factors = oracle.factors_by_name()
        factor = factors["false the Java API was an original work"]
        assert factor.terms[0].name == "the Java API"


class TestImplication:
    def test_opinion_implies_holding_group(self, make_opinion_with_holding):
        """Can be affected by order of tests."""
        watt = make_opinion_with_holding["watt_majority"]
        holdings = watt.holdings[:2]
        assert isinstance(holdings, HoldingGroup)
        assert watt.implies(holdings)
        explanation = watt.explain_implication(holdings)
        assert len(explanation.reasons) == 2

    def test_opinion_implied_by_rule(
        self, make_opinion_with_holding, make_holding, make_rule
    ):
        """Can be affected by order of tests."""
        watt = make_opinion_with_holding["oracle_majority"]
        watt.clear_holdings()
        watt.posit(make_holding["h2"])
        assert watt.implied_by(make_rule["h2_despite_due_process"])
        assert not watt.implies(make_rule["h2_despite_due_process"])

    def test_no_implication(self, make_opinion_with_holding):
        watt = make_opinion_with_holding["watt_majority"]
        brad = make_opinion_with_holding["brad_majority"]
        assert not watt >= brad
        assert not watt.implies(brad.holdings)

    def test_posit_list_of_holdings_and_imply(
        self, make_opinion_with_holding, make_response
    ):
        watt = make_opinion_with_holding["watt_majority"]
        brad = make_opinion_with_holding["brad_majority"]
        watt.clear_holdings()
        brad.clear_holdings()
        client = FakeClient(responses=make_response)
        some_rules_raw = loaders.load_holdings(filename="holding_watt.yaml")
        some_rules = readers.read_holdings(some_rules_raw, client=client)
        for case in (watt, brad):
            case.clear_holdings()
            case.posit(some_rules[:3])
        watt.posit(some_rules[3])
        assert watt > brad
        assert not brad >= watt

    def test_error_to_compare_to_str(self, make_opinion_with_holding):
        watt = make_opinion_with_holding["watt_majority"]
        with pytest.raises(TypeError):
            watt.implies("this")

    def test_error_to_check_contradiction_of_str(self, make_opinion_with_holding):
        watt = make_opinion_with_holding["watt_majority"]
        with pytest.raises(TypeError):
            watt.contradicts("this")

    def test_opinion_implies_None(self, make_opinion_with_holding, make_holding):
        watt = make_opinion_with_holding["watt_majority"]
        assert watt.implies(None)

    def test_opinion_implies_holding(self, make_opinion_with_holding, make_holding):
        watt = make_opinion_with_holding["watt_majority"]
        watt.clear_holdings()
        watt.posit(make_holding["h2_invalid_undecided"])
        assert watt >= make_holding["h2_undecided"]
        assert watt > make_holding["h2_undecided"]

    def test_opinion_does_not_imply_holding(
        self, make_opinion_with_holding, make_holding
    ):
        watt = make_opinion_with_holding["watt_majority"]
        watt.clear_holdings()
        watt.posit(make_holding["h2_irrelevant_inputs_undecided"])
        assert not watt >= make_holding["h2_undecided"]
        assert not watt > make_holding["h2_undecided"]

    def test_opinion_does_not_imply_holding_group(self, make_opinion_with_holding):
        watt = make_opinion_with_holding["watt_majority"]
        holdings = make_opinion_with_holding["oracle_majority"].holdings
        assert isinstance(holdings, HoldingGroup)
        assert not watt.implies(holdings)
        explanation = watt.explain_implication(holdings)
        assert explanation is None

    def test_opinion_implied_by_decision(
        self, make_holding, make_decision_with_holding
    ):
        watt = make_decision_with_holding["watt"]
        reading = OpinionReading()
        reading.posit(watt.holdings[0])
        assert reading.implied_by(watt)


class TestContradiction:
    def test_opinion_contradicts_opinion(self, make_opinion_with_holding):
        """Return the only contradictory pair of Holdings between these two Opinions."""
        oracle = make_opinion_with_holding["oracle_majority"]
        lotus = make_opinion_with_holding["lotus_majority"]

        left = OpinionReading(
            opinion_type="majority", opinion_author=oracle.opinion_author
        )
        left.posit(lotus.holdings[6])
        right = OpinionReading(
            opinion_type="majority", opinion_author=lotus.opinion_author
        )
        right.posit(oracle.holdings[10])
        explanation = left.explain_contradiction(right)
        assert len(explanation.reasons) == 1
        assert isinstance(explanation.reasons[0].left, Holding)
        assert isinstance(explanation.reasons[0].right, Holding)

    def test_contradiction_of_holding(
        self, make_opinion_with_holding, e_search_clause, make_holding
    ):
        assert make_opinion_with_holding["watt_majority"].contradicts(
            make_holding["h2_output_false_ALL_MUST"] + e_search_clause
        )

    def test_explain_opinion_contradicting_holding(self, make_opinion_with_holding):
        oracle = make_opinion_with_holding["oracle_majority"]
        lotus = make_opinion_with_holding["lotus_majority"]
        explanation = oracle.explain_contradiction(lotus.holdings[6])
        assert (
            "<the java api> is like <the lotus menu command hierarchy>"
            in str(explanation).lower()
        )

    def test_explain_opinion_contradicting_rule(self, make_opinion_with_holding):
        oracle = make_opinion_with_holding["oracle_majority"]
        lotus = make_opinion_with_holding["lotus_majority"]
        explanation = oracle.explain_contradiction(lotus.holdings[6].rule)
        expected = "<the java api> is like <the lotus menu command hierarchy>"
        assert expected in str(explanation).lower()

    def test_contradiction_of_decision(
        self, make_opinion_with_holding, make_decision_with_holding
    ):
        assert make_opinion_with_holding["oracle_majority"].contradicts(
            make_decision_with_holding["lotus"]
        )

    def test_explain_opinion_contradicting_decision(
        self, make_opinion_with_holding, make_decision_with_holding
    ):
        oracle_majority = make_opinion_with_holding["oracle_majority"]
        lotus = make_decision_with_holding["lotus"]
        explanation = oracle_majority.explain_contradiction(lotus)
        assert "contradicts" in str(explanation).lower()

    def test_no_explanation_of_contradiction(self, make_opinion_with_holding):
        watt = make_opinion_with_holding["watt_majority"]
        holdings = watt.holdings[:2]
        assert isinstance(holdings, HoldingGroup)
        assert watt.implies(holdings)
        explanation = watt.explain_contradiction(holdings)
        assert explanation is None

    def test_error_contradiction_with_procedure(
        self, make_opinion_with_holding, make_procedure
    ):
        with pytest.raises(TypeError):
            make_opinion_with_holding["watt_majority"].contradicts(make_procedure["c1"])



"""AuthoritySpoke: Reading the law for the last time."""

from justopinion.decisions import Decision
from justopinion import CAPClient, Opinion
from legislice import Enactment
from legislice.download import Client as LegisClient
from nettlesome.entities import Entity
from nettlesome.predicates import Predicate
from nettlesome.quantities import Comparison

from .decisions import DecisionReading
from .facts import Fact, Exhibit, Evidence, Allegation, Pleading
from .holdings import Holding
from .opinions import OpinionReading
from .rules import Rule

__version__ = "0.9.1"



r""":class:`Court` documents that decide litigation and posit :class:`.Rule`\s."""

from __future__ import annotations

from itertools import zip_longest
from typing import Any, Dict, Iterator, List
from typing import Optional, Sequence, Union

import logging

from anchorpoint.textselectors import (
    TextQuoteSelector,
    TextPositionSet,
    TextPositionSelector,
)
from justopinion import Opinion
from legislice.enactments import EnactmentPassage
from nettlesome.terms import Comparable, ContextRegister, Explanation, Term
from nettlesome.factors import Factor
from pydantic import BaseModel, validator

from authorityspoke.facts import Entity, Fact, Allegation, Pleading, Exhibit, Evidence
from authorityspoke.holdings import Holding, HoldingGroup
from authorityspoke.procedures import Procedure
from authorityspoke.rules import Rule


logger = logging.getLogger(__name__)


class EnactmentWithAnchors(BaseModel):
    """A term with a set of anchors."""

    passage: EnactmentPassage
    anchors: TextPositionSet = TextPositionSet()

    @validator("anchors", pre=True)
    def validate_anchors(cls, value: TextPositionSet) -> TextPositionSet:
        """Validate that the anchors are non-empty."""

        if value is None:
            return TextPositionSet()
        return value


class TermWithAnchors(BaseModel):
    """A term with a set of anchors."""

    term: Union[Entity, Fact, Allegation, Pleading, Exhibit, Evidence]
    anchors: TextPositionSet = TextPositionSet()

    @validator("anchors", pre=True)
    def validate_anchors(cls, value: TextPositionSet) -> TextPositionSet:
        """Validate that the anchors are non-empty."""

        if value is None:
            return TextPositionSet()
        return value


class HoldingWithAnchors(BaseModel):
    """A :class:`.Holding` with a :class:`.TextPositionSet` that anchors it."""

    holding: Holding
    anchors: TextPositionSet = TextPositionSet()

    @validator("anchors", pre=True)
    def validate_anchors(cls, value: TextPositionSet) -> TextPositionSet:
        """Validate that the anchors are non-empty."""
        if value is None:
            return TextPositionSet()
        return value


class AnchoredHoldings(BaseModel):
    """Holdings with objects storing the Holdings' links to Opinion text."""

    holdings: List[HoldingWithAnchors] = []
    named_anchors: List[TermWithAnchors] = []
    enactment_anchors: List[EnactmentWithAnchors] = []

    def find_term_index(self, term: Term) -> Optional[int]:
        """Find the index of a term in the holdings."""
        key = term.key
        for index, term_with_anchors in enumerate(self.named_anchors):
            if term_with_anchors.term.key == key:
                return index
        return None

    def add_term(self, term: Term, anchors: TextPositionSet) -> None:
        """Add a term that can be found in self's holdings, with the term's anchors to the text."""
        term_index = self.find_term_index(term)
        if term_index is None:
            self.named_anchors.append(TermWithAnchors(term=term, anchors=anchors))
        else:
            self.named_anchors[term_index].anchors += anchors

    def get_term_anchors(self, key: str) -> TextPositionSet:
        """Get the anchors for a term."""
        for item in self.named_anchors:
            if item.term.key == key:
                return item.anchors
        raise KeyError(f"Term with key '{key}' not found")

    def find_enactment_index(self, enactment: EnactmentPassage) -> Optional[int]:
        """Find the index of a term in the holdings."""
        key = str(enactment)
        for index, enactment_with_anchors in enumerate(self.enactment_anchors):
            if str(enactment_with_anchors.passage) == key:
                return index
        return None

    def add_enactment(
        self, enactment: EnactmentPassage, anchors: TextPositionSet
    ) -> None:
        """Add EnactmentPassage with text anchors, if it isn't a duplicate."""
        term_index = self.find_enactment_index(enactment)
        if term_index is None:
            self.enactment_anchors.append(
                EnactmentWithAnchors(passage=enactment, anchors=anchors)
            )
        else:
            self.enactment_anchors[term_index].anchors += anchors

    def get_enactment_anchors(self, key: str) -> TextPositionSet:
        """Get the anchors for a term."""
        for item in self.enactment_anchors:
            if str(item.passage) == key:
                return item.anchors
        raise KeyError(f"Enactment passage with key '{key}' not found")


class OpinionReading(Comparable, BaseModel):
    """An interpretation of what Holdings are supported by the text of an Opinion."""

    anchored_holdings: AnchoredHoldings = AnchoredHoldings()
    opinion_type: str = ""
    opinion_author: str = ""

    @property
    def holdings(self) -> HoldingGroup:
        """Get Holdings for the Opinion."""
        return HoldingGroup([item.holding for item in self.anchored_holdings.holdings])

    @property
    def holding_anchors(self) -> List[TextPositionSet]:
        """Get Holdings with corresponding anchors for the Opinion."""
        return [item.anchors for item in self.anchored_holdings.holdings]

    @property
    def anchored_factors(self) -> List[TermWithAnchors]:
        """Get Factors with corresponding anchors for the Opinion."""
        return self.anchored_holdings.named_anchors

    @property
    def anchored_enactments(self) -> List[EnactmentWithAnchors]:
        """Get Enactment passages with corresponding anchors for the Opinion."""
        return self.anchored_holdings.enactment_anchors

    def __str__(self):
        return super().__str__()

    def factors(self) -> List[Factor]:
        """Get all factors from the Holdings of the Opinion, without duplication."""
        factors_by_name = self.factors_by_name()
        return list(factors_by_name.values())

    def factors_by_name(self) -> FactorIndex:
        """Get an index of Factors, indexed by name."""
        factor_index = FactorIndex()
        for holding in self.holdings:
            factors_for_holding = holding.recursive_terms
            for value in factors_for_holding.values():
                if not isinstance(value, Holding):
                    factor_index.insert_by_name(value=value)
        return factor_index

    def clear_holdings(self):
        r"""Remove all :class:`.Holding`\s from the opinion."""
        self.anchored_holdings.holdings = []

    def get_enactment_anchors(self, key: str) -> TextPositionSet:
        """Get the anchors for an enactment passage."""
        return self.anchored_holdings.get_enactment_anchors(key)

    def get_term_anchors(self, key: str) -> TextPositionSet:
        """Get the anchors for an enactment passage."""
        return self.anchored_holdings.get_term_anchors(key)

    def explanations_contradiction(
        self,
        other: Comparable,
    ) -> Iterator[Explanation]:
        """Yield contexts that would result in a contradiction between self and other."""
        if not self.comparable_with(other):
            raise TypeError(
                f"'Implies' test not implemented for types {self.__class__} and {other.__class__}."
            )
        if isinstance(other, Rule):
            other = Holding(rule=other)
        if isinstance(other, Holding):
            yield from self.holdings.explanations_contradiction(other)

        if isinstance(other, HoldingGroup):
            yield from self.holdings.explanations_contradiction(other)

        elif isinstance(other, self.__class__):
            yield from self.holdings.explanations_contradiction(other.holdings)

        elif hasattr(other, "explanations_contradiction"):
            yield from other.explanations_contradiction(self)

    def comparable_with(self, other: Any) -> bool:
        """Check if other can be compared to self for implication or contradiction."""
        if not isinstance(other, Comparable):
            return False
        if isinstance(other, Procedure):
            return False
        return not isinstance(other, Factor)

    def contradicts(
        self, other: Comparable, context: Optional[ContextRegister] = None
    ) -> bool:
        """Check if other contradicts one of self's Holdings."""
        if not self.comparable_with(other):
            raise TypeError(
                "'Contradicts' test not implemented for types "
                + f"{self.__class__} and {other.__class__}."
            )
        return any(
            explanation is not None
            for explanation in self.explanations_contradiction(other)
        )

    def implies(
        self, other: Optional[Comparable], context: Optional[ContextRegister] = None
    ) -> bool:
        """Check if all of other's Holdings are implied by holdings of self."""
        if other is None:
            return True
        return any(self.explanations_implication(other))

    def explain_implication(
        self,
        other: Comparable,
    ) -> Optional[Explanation]:
        """Get first Explanation of how one of self's Holdings implies other."""
        explanations = self.explanations_implication(other)
        try:
            explanation = next(explanations)
        except StopIteration:
            return None
        return explanation

    def explain_contradiction(self, other: Comparable) -> Optional[Explanation]:
        """Get first Explanation of how other contradicts one of self's Holdings."""
        explanations = self.explanations_contradiction(other)
        try:
            explanation = next(explanations)
        except StopIteration:
            return None
        return explanation

    def explanations_implication(
        self,
        other: Comparable,
    ) -> Iterator[Union[ContextRegister, Explanation]]:
        """Yield contexts that would result in self implying other."""
        if not self.comparable_with(other):
            raise TypeError(
                f"'Implies' test not implemented for types {self.__class__} and {other.__class__}."
            )
        if isinstance(other, Rule):
            other = Holding(rule=other)
        if isinstance(other, Holding):
            for self_holding in self.holdings:
                yield from self_holding.explanations_implication(other)
        if isinstance(other, HoldingGroup):
            yield from self.holdings.explanations_implication(other)
        elif isinstance(other, self.__class__):
            yield from self.holdings.explanations_implication(other.holdings)
        else:
            yield from other.explanations_implied_by(self)

    def generic_terms_by_str(self) -> Dict[str, Comparable]:
        r"""
        Get all generic :class:`.Factor`\s mentioned in ``self``.

        :returns:
            a list of generic :class:`.Factor` objects mentioned in
            any ``input``, ``output``, or ``despite`` :class:`.Factor`
            of ``self``, with guaranteed order, including each
            generic :class:`.Factor` only once.
        """
        generics: Dict[str, Comparable] = {}
        for holding in self.holdings:
            for generic in holding.generic_terms():
                generics[str(generic)] = generic
        return generics

    def get_factor_by_name(self, name: str) -> Optional[Comparable]:
        """
        Search recursively in holdings of ``self`` for :class:`.Factor` with ``name``.

        :param name:
            string to match with the ``name`` attribute of the
            :class:`.Factor` being searched for.

        :returns:
            a :class:`.Factor` with the specified ``name``, if one
            exists in a :class:`.Rule` in ``self.holdings``.
            Otherwise ``None``.
        """

        for holding in self.holdings:
            factor = holding.get_factor_by_name(name)
            if factor is not None:
                return factor
        return None

    def get_factor_by_str(self, query: str) -> Optional[Factor]:
        r"""Search recursively for :class:`.Factor` in holdings of self."""
        return self.holdings.get_factor_by_str(query)

    def get_matching_holding(self, holding: Holding) -> Optional[Holding]:
        """Check self's Holdings for a Holding with the same meaning."""
        for known_holding in self.holdings:
            if holding.means(known_holding):
                return known_holding
        return None

    def posit_holding(
        self,
        holding: Union[Holding, Rule, HoldingWithAnchors],
        holding_anchors: Optional[
            Union[TextPositionSelector, TextQuoteSelector, TextPositionSet]
        ] = None,
        named_anchors: Optional[List[TermWithAnchors]] = None,
        enactment_anchors: Optional[List[EnactmentWithAnchors]] = None,
        context: Optional[Sequence[Factor]] = None,
    ) -> None:
        r"""Record that this Opinion endorses specified :class:`Holding`\s."""
        named_anchors = named_anchors or []
        enactment_anchors = enactment_anchors or []

        if isinstance(holding, HoldingWithAnchors):
            holding, holding_anchors = holding.holding, holding.anchors
        if isinstance(holding_anchors, (TextQuoteSelector, str)):
            holding_anchors = [holding_anchors]
        if isinstance(holding_anchors, List) and isinstance(
            holding_anchors[0], (str, TextQuoteSelector)
        ):
            holding_anchors = TextPositionSet.from_quotes(holding_anchors)
        if isinstance(holding, Rule):
            logger.warning(
                "posit_holding was called with a Rule "
                "that was automatically converted to a Holding"
            )
            holding = Holding(rule=holding)

        if not isinstance(holding, Holding):
            raise TypeError('"holding" must be an object of type Holding.')

        for named_anchor in named_anchors:
            self.anchored_holdings.add_term(
                term=named_anchor.term, anchors=named_anchor.anchors
            )

        for enactment_anchor in enactment_anchors:
            self.anchored_holdings.add_enactment(
                enactment=enactment_anchor.passage, anchors=named_anchor.anchors
            )

        matching_holding = self.get_matching_holding(holding)
        if matching_holding:
            matching_holding.anchors += holding.anchors
        else:
            if context:
                holding = holding.new_context(context, source=self)
            self.anchored_holdings.holdings.append(
                HoldingWithAnchors(holding=holding, anchors=holding_anchors)
            )

    def posit_holdings(
        self,
        holdings: Union[
            AnchoredHoldings, List[Union[HoldingWithAnchors, Holding, Rule]]
        ],
        holding_anchors: Optional[List[HoldingWithAnchors]] = None,
        named_anchors: Optional[List[TermWithAnchors]] = None,
        enactment_anchors: Optional[List[EnactmentWithAnchors]] = None,
        context: Optional[Sequence[Factor]] = None,
    ):
        r"""
        Add :class:`.Holding`\s to this ``OpinionReading`` from a sequence.

        :param holdings:
            a sequence of :class:`.Holding`\s that this :class:`.OpinionReading`
            posits as valid in its own court or jurisdiction, regardless of
            whether ``self`` accepts that the ``inputs`` correspond to the
            reality of the current case, and regardless of whether the
            court orders that the ``outputs`` be put into effect.

        :param context:
            an ordered sequence (probably :py:class:`dict`) of
            generic :class:`.Factor` objects from ``self`` which
            will provide the context for the new holding in the
            present case.
        """
        if isinstance(holdings, AnchoredHoldings):
            for holding_with_anchor in holdings.holdings:
                self.posit_holding(
                    holding=holding_with_anchor,
                    named_anchors=named_anchors,
                    enactment_anchors=enactment_anchors,
                    context=context,
                )
        else:
            holding_anchors = holding_anchors or []
            for (holding, selector_list) in zip_longest(holdings, holding_anchors):
                self.posit_holding(
                    holding=holding,
                    holding_anchors=selector_list,
                    named_anchors=named_anchors,
                    enactment_anchors=enactment_anchors,
                    context=context,
                )

    def posit(
        self,
        holdings: Union[
            AnchoredHoldings,
            Holding,
            Rule,
            HoldingWithAnchors,
            List[Union[HoldingWithAnchors, Holding, Rule]],
        ],
        holding_anchors: Optional[List[HoldingWithAnchors]] = None,
        named_anchors: Optional[List[TermWithAnchors]] = None,
        enactment_anchors: Optional[List[EnactmentWithAnchors]] = None,
        context: Optional[Sequence[Factor]] = None,
    ) -> None:
        r"""
        Add one or more :class:`.Holding`\s to this ``OpinionReading``.

        This method passes its values to :meth:`~posit_holding` or
        :meth:`~posit_holdings` depending on whether the `holding` parameter
        is one :class:`.Holding` or a :class:`list`\.

        :param holdings:
            a :class:`.Holding` that the :class:`.OpinionReading` ``self`` posits
            as valid in its own court or jurisdiction, regardless of
            whether ``self`` accepts that the ``inputs`` of the
            :class:`.Holding` correspond to the reality of the current
            case, and regardless of whether the court orders that
            the ``outputs`` of the :class:`.Holding` be put into effect.

        :param context:
            an ordered sequence (probably :py:class:`dict`) of
            generic :class:`.Factor` objects from ``self`` which
            will provide the context for the new holding in the
            present case.
        """
        if isinstance(holdings, (Holding, Rule, HoldingWithAnchors)):
            self.posit_holding(
                holdings,
                holding_anchors=holding_anchors,
                named_anchors=named_anchors,
                enactment_anchors=enactment_anchors,
                context=context,
            )
        else:
            self.posit_holdings(
                holdings,
                holding_anchors=holding_anchors,
                named_anchors=named_anchors,
                enactment_anchors=enactment_anchors,
                context=context,
            )

    def _implied_by_holding(self, other: Holding, context: Explanation) -> bool:
        return all(
            other.implies(self_holding, context=context.reversed_context())
            for self_holding in self.holdings
        )

    def _implied_by_rule(self, other: Rule, context: Explanation) -> bool:
        return self._implied_by_holding(other=Holding(rule=other), context=context)

    def implied_by(
        self,
        other: Union[OpinionReading, Holding, Rule],
        context: Optional[Union[ContextRegister, Explanation]] = None,
    ) -> bool:
        """Determine if other implies all the Holdings of self."""
        if not isinstance(context, Explanation):
            context = Explanation.from_context(context)
        if isinstance(other, Holding):
            return self._implied_by_holding(other, context=context)
        elif isinstance(other, Rule):
            return self._implied_by_rule(other, context=context)
        return other.implies(self, context=context.reversed_context())

    def __ge__(self, other: Union[OpinionReading, Holding, Rule]) -> bool:
        """
        Find whether ``self``'s holdings imply all the holdings of ``other``.

        :returns:
            a bool indicating whether the :class:`.Holding` ``other``
            (or every holding of ``other``, if other is an :class:`.OpinionReading`)
            is implied by some :class:`.Holding` in ``self.holdings``.
        """
        return self.implies(other)

    def __gt__(self, other) -> bool:
        """
        Find whether ``self``\'s holdings imply ``other``\'s but self != other.

        This actually tests for inequality because ``OpinionReading`` does not
        have a ``means`` method.

        :returns:
            self >= other and self != other.
        """
        return (self >= other) and (self != other)


class FactorIndex(Dict[str, Factor]):
    r"""Index of :class:`.Factor`/s that may share common anchors."""

    def insert_by_name(self, value: Factor) -> None:
        """
        Insert Factor using its name as key if possible.

        If the Factor has no name attr, use its str as key instead.
        """
        if value.name:
            self.insert(key=value.name, value=value)
            return None
        key = str(value)
        self.insert(key=key, value=value)

    def insert(self, key: str, value: Factor) -> None:
        """Insert Factor using its str as its key."""
        if key in self.keys():
            if value.name:
                if not self[key].name:
                    self[key].name = value.name
                if value.name != self[key].name:
                    raise NameError(
                        f"{type(value)} objects with identical representation ({str(value)}) "
                        f"have differing names: '{value.name}' and '{self[key].name}'"
                    )
        else:
            self[key] = value



"""
Converting simple structured data from XML or JSON into authorityspoke objects.

These functions will usually be called by functions from the io.loaders module
after they import some data from a file.
"""
from typing import Any, NamedTuple
from typing import Dict, List, Optional, Tuple, Sequence, Union


from anchorpoint.textselectors import TextQuoteSelector
from legislice.download import Client
from legislice.types import RawEnactment
from nettlesome.entities import Entity

from authorityspoke.decisions import Decision, DecisionReading, RawDecision
from authorityspoke.facts import Fact, Exhibit, Evidence, Allegation, Pleading
from authorityspoke.holdings import Holding, RawHolding
from authorityspoke.opinions import (
    AnchoredHoldings,
    EnactmentWithAnchors,
    TermWithAnchors,
    HoldingWithAnchors,
)
from authorityspoke.facts import RawFactor
from authorityspoke.io.name_index import index_names, Mentioned, collect_enactments
from authorityspoke.io.text_expansion import expand_shorthand

RawSelector = Union[str, Dict[str, str]]


FACTOR_SUBCLASSES = {
    class_obj.__name__: class_obj
    for class_obj in (Allegation, Entity, Exhibit, Evidence, Fact, Pleading)
}


class HoldingsIndexed(NamedTuple):
    """Lists :class:`.Holding` objects with corresponding text selectors."""

    holdings: List[Holding]
    mentioned: Mentioned
    holding_anchors: List[List[TextQuoteSelector]]


def collect_anchors_from_index(object_index, field_name: str):
    """Get text anchors out of an index of terms or enactments."""
    result = []
    for key, value in object_index.items():
        if value.get("anchors"):
            anchored_object: Dict[str, Any] = {}
            anchors = value.pop("anchors")
            if isinstance(anchors, List):
                anchors = [anchor for anchor in anchors if anchor != "|"]
            anchored_object["anchors"] = anchors
            anchored_object[field_name] = value
            result.append(anchored_object)
    return result, object_index


def read_holdings_with_anchors(
    record: Dict[str, Union[List[RawHolding], List[RawSelector]]],
    client: Optional[Client] = None,
) -> AnchoredHoldings:
    r"""
    Load a list of Holdings from JSON, with text links.

    :param record:
        a list of dicts representing holdings, in the JSON input format

    :param client:
        Legislice client for downloading missing fields from `record`

    :returns:
        a namedtuple listing :class:`.Holding` objects with
        a list matching :class:`.Holding`\s to selectors and
        an index matching :class:`.Factor`\s to selectors.
    """

    (
        holdings,
        enactment_anchors,
        factor_anchors,
        holding_anchors,
    ) = extract_anchors_from_holding_record(record, client)
    holdings_with_anchors = []
    for i, holding in enumerate(holdings):
        new = HoldingWithAnchors(holding=holding, anchors=holding_anchors[i])
        holdings_with_anchors.append(new)
    return AnchoredHoldings(
        holdings=holdings_with_anchors,
        named_anchors=factor_anchors,
        enactment_anchors=enactment_anchors,
    )


def expand_factor(
    record: Union[str, RawFactor], factor_index: Mentioned
) -> Union[str, RawFactor]:
    """Expand fields of Factor from index of mentioned factors."""
    to_expand = [
        "statement",
        "statement_attribution",
        "fact",
        "offered_by",
        "exhibit",
        "to_effect",
        "filer",
        "pleading",
    ]
    expanded = (
        factor_index.get_if_present(record) if isinstance(record, str) else record
    )
    if not isinstance(expanded, Dict):
        return expanded
    if "terms" in expanded:
        expanded["terms"] = expand_names(expanded["terms"], factor_index)
    for field in to_expand:
        if field in expanded:
            expanded[field] = expand_factor(expanded[field], factor_index)
    return expanded


def expand_names(
    record: List[Union[str, RawFactor]], factor_index: Mentioned
) -> List[RawFactor]:
    r"""Expand a list of names into a list of factors."""
    if isinstance(record, str):
        record = [record]
    if isinstance(record, bool):
        return record

    result = []

    for name in record:

        result.append(expand_factor(name, factor_index=factor_index))

    return result


def expand_enactments(
    record: List[Union[str, RawEnactment]], enactment_index: Mentioned
) -> List[RawEnactment]:
    r"""
    Expand a list of enactments into a list of dicts.

    :param record:
        a list of enactments, either as strings or dicts

    :param enactment_index:
        a dict of names to enactments

    :returns:
        a list of dicts representing enactments
    """
    return [enactment_index.get_if_present(name) for name in record]


def walk_tree_and_expand(
    obj: Union[Dict, List], mentioned: Mentioned, ignore: Sequence[str] = ()
) -> Union[Dict, List]:
    """
    Traverse tree of dicts and lists, and modify each node.

    :param obj: the object to traverse

    :param func:
        the function to call on each dict node, returning a dict

    :param ignore: the names of keys that should not be explored

    :returns: a version of the tree with every node modified by `func`
    """
    if isinstance(obj, str):
        obj = mentioned.get_if_present(obj)
    if isinstance(obj, List):
        obj = [mentioned.get_if_present(item) for item in obj]
        return [walk_tree_and_expand(item, mentioned, ignore) for item in obj]
    if isinstance(obj, Dict):

        obj_dict: Dict = {}
        for key, value in obj.items():
            if key not in ignore:
                obj_dict[key] = mentioned.get_if_present(value)
            else:
                obj_dict[key] = value

        for key, value in obj_dict.items():
            if isinstance(value, (Dict, List)) and key not in ignore:
                obj_dict[key] = walk_tree_and_expand(value, mentioned, ignore)

        return obj_dict

    return obj


def expand_holding(
    record: RawHolding, factor_index: Mentioned, enactment_index: Mentioned
) -> RawHolding:
    """Expand one holding from index of expanded terms and enactments."""
    new_index = Mentioned({**factor_index, **enactment_index})
    return walk_tree_and_expand(
        record,
        mentioned=new_index,
        ignore=["predicate", "enactment", "selection", "name"],
    )


def expand_holdings(
    record: List[Union[str, RawHolding]],
    factor_index: Mentioned,
    enactment_index: Mentioned,
) -> List[RawHolding]:
    """Expand holdings from index of expanded terms and enactments."""
    if isinstance(record, dict):
        record = [record]
    holdings = [factor_index.get_if_present(holding) for holding in record]
    holdings = [
        expand_holding(
            holding, factor_index=factor_index, enactment_index=enactment_index
        )
        for holding in holdings
    ]
    return holdings


def extract_anchors_from_holding_record(
    record: List[RawHolding], client: Optional[Client] = None
) -> Tuple[
    List[Holding],
    List[EnactmentWithAnchors],
    List[TermWithAnchors],
    List[Dict[str, str]],
]:
    r"""
    Load a list of Holdings from JSON, with text links.

    :param record:
        a list of dicts representing holdings, in the JSON input format

    :param client:
        Legislice client for downloading missing fields from `record`

    :returns:
        a tuple of four objects containing holdings, terms, enactments,
        and anchors.
    """
    record_post_enactments, enactment_index = collect_enactments(record)
    if client:
        enactment_index_post_client = client.update_entries_in_enactment_index(
            enactment_index
        )
    else:
        enactment_index_post_client = enactment_index
    enactment_anchors, enactment_index_post_anchors = collect_anchors_from_index(
        enactment_index_post_client, "passage"
    )

    enactment_result = []
    for anchor in enactment_anchors:
        anchor["passage"] = enactment_index_post_anchors.get_if_present(
            anchor["passage"]
        )
        enactment_result.append(EnactmentWithAnchors(**anchor))

    record_post_terms, factor_index = index_names(record_post_enactments)
    factor_anchors, factor_index_post_anchors = collect_anchors_from_index(
        factor_index, "term"
    )

    factor_result = []
    for anchor in factor_anchors:
        anchor["term"] = expand_holding(
            anchor["term"],
            factor_index=factor_index_post_anchors,
            enactment_index=enactment_index_post_anchors,
        )
        factor_result.append(TermWithAnchors(**anchor))

    factor_anchors = [TermWithAnchors(**anchor) for anchor in factor_anchors]

    expanded = expand_holdings(
        record_post_terms,
        factor_index=factor_index_post_anchors,
        enactment_index=enactment_index_post_anchors,
    )
    holding_anchors = [holding.pop("anchors", None) for holding in expanded]

    result = []
    for holding in expanded:
        result.append(Holding(**holding))
    return result, enactment_result, factor_result, holding_anchors


def read_holdings(
    record: List[RawHolding], client: Optional[Client] = None
) -> List[Holding]:
    r"""
    Load a list of :class:`Holdings`\s from JSON.

    :param record:
        a list of dicts representing holdings, in the JSON input format

    :parame regime:
        A collection of :class:`.Jurisdiction`\s and the :class:`.Code`\s
        that have been enacted in each. Used for constructing
        :class:`.Enactment`\s referenced by :class:`.Holding`\s.

    :returns:
        a list of :class:`.Holding` objects
    """
    (
        holdings,
        enactment_anchors,
        factor_anchors,
        holding_anchors,
    ) = extract_anchors_from_holding_record(record, client)

    return holdings


def read_decision(decision: Union[RawDecision, Decision]) -> DecisionReading:
    r"""
    Create and return a :class:`~authorityspoke.decisions.Decision` from a dict API response.

    Relies on the JSON format from the `Caselaw Access Project
    API <https://api.case.law/v1/cases/>`_.

    :param decision_dict:
        A dict created from a Caselaw Access Project API response.
    """
    if not isinstance(decision, Decision):
        decision = Decision(**decision)
    return DecisionReading(decision=decision)



import setuptools

with open("README.rst", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="AuthoritySpoke",
    version="0.9.1",
    author="Matthew Carey",
    author_email="matt@authorityspoke.com",
    description="legal authority automation",
    long_description=long_description,
    long_description_content_type="text/x-rst",
    url="https://github.com/mscarey/authorityspoke",
    project_urls={
        "Bug Tracker": "https://github.com/mscarey/authorityspoke/issues",
        "Documentation": "https://authorityspoke.readthedocs.io/en/latest/",
    },
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Legal Industry",
        "License :: Free To Use But Restricted",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Operating System :: OS Independent",
        "Natural Language :: English",
    ],
    packages=setuptools.find_packages(exclude=["tests"]),
    install_requires=[
        "lxml==4.8.0",
        "anchorpoint~=0.7.0",
        "eyecite~=2.3.4",
        "justopinion~=0.2.5",
        "legislice~=0.7.0",
        "nettlesome~=0.6.1",
        "pint>=0.15",
        "pydantic",
        "python-dotenv",
        "python-ranges~=0.2.1",
        "python-slugify",
        "pyyaml",
        "requests",
        "roman",
        "sympy>=1.7.1",
    ],
    python_requires=">=3.8",
)



"""Create models of assertions accepted as factual by courts."""
from __future__ import annotations
from copy import deepcopy
import operator
from typing import Any, ClassVar, Dict, Iterator, List
from typing import Mapping, Optional, Sequence, Tuple, Union

from pydantic import BaseModel, Extra, ValidationError, validator, root_validator
from slugify import slugify

from nettlesome.entities import Entity
from nettlesome.factors import Factor
from nettlesome.formatting import indented, wrapped
from nettlesome.terms import (
    Comparable,
    ContextRegister,
    Explanation,
    Term,
    TermSequence,
    new_context_helper,
)
from nettlesome.predicates import Predicate
from nettlesome.quantities import Comparison, QuantityRange


RawPredicate = Dict[str, Union[str, bool]]
RawFactor = Dict[str, Union[RawPredicate, Sequence[Any], str, bool]]


class Fact(Factor, BaseModel):
    r"""
    An assertion accepted as factual by a court.

    Often based on factfinding by a judge or jury.

    Facts may be "alleged" by a pleading, "supported" by evidence, or
    "found" to be factual by a jury verdict or a judge's finding of fact.

    :param predicate:
        a natural-language clause with zero or more slots
        to insert ``terms`` that are typically the
        subject and objects of the clause.

    :param terms:
        a series of :class:`Factor` objects that fill in
        the blank spaces in the ``predicate`` statement.

    :param name:
        an identifier for this object, often used if the object needs
        to be referred to multiple times in the process of composing
        other :class:`Factor` objects.

    :param standard_of_proof:
        a descriptor for the degree of certainty associated
        with the assertion in the ``predicate``.

    :param absent:
        whether the absence, rather than the presence, of the legal
        fact described above is being asserted.

    :param generic:
        whether this object could be replaced by another generic
        object of the same class without changing the truth of the
        :class:`Rule` in which it is mentioned.

    :attr standards_of_proof:
        a tuple with every allowable name for a standard of
        proof, in order from weakest to strongest.

        .. note:
            If any courts anywhere in a legal regime disagree about the
            relative strength of the various standards of proof, or if
            any court considers the order context-specific, then this
            approach of hard-coding their names and order will have to change.
    """

    predicate: Union[Predicate, Comparison]
    terms: List[
        Union[Entity, "Fact", "Allegation", "Pleading", "Exhibit", "Evidence"]
    ] = []
    name: str = ""
    absent: bool = False
    generic: bool = False
    standard_of_proof: Optional[str] = None
    standards_of_proof: ClassVar[Tuple[str, ...]] = (
        "scintilla of evidence",
        "substantial evidence",
        "preponderance of evidence",
        "clear and convincing",
        "beyond reasonable doubt",
    )

    class Config:
        """Fail validation if the input data has data not specified in the model."""

        extra = Extra.forbid

    @root_validator(pre=True)
    def nest_predicate_fields(cls, values):
        """Move fields passed to the Fact model that really belong to the Predicate model."""
        type_str = values.pop("type", "")
        if type_str and type_str.lower() != "fact":
            raise ValidationError(f"type {type_str} was passed to Fact model")

        if isinstance(values.get("predicate"), str):
            values["predicate"] = Predicate(content=values["predicate"])
            if "truth" in values:
                values["predicate"].truth = values.pop("truth")

        for field_name in ["content", "truth", "sign", "expression"]:
            if field_name in values:
                values["predicate"] = values.get("predicate", {})
                values["predicate"][field_name] = values.pop(field_name)
        if isinstance(values.get("predicate"), dict) and values["predicate"].get(
            "content"
        ):
            for sign in {
                **QuantityRange.opposite_comparisons,
                **QuantityRange.normalized_comparisons,
            }:
                if sign in values["predicate"]["content"]:
                    content, quantity_text = values["predicate"]["content"].split(sign)
                    values["predicate"]["content"] = content.strip()
                    values["predicate"]["expression"] = quantity_text.strip()
                    values["predicate"]["sign"] = sign
                    break
        return values

    @validator("terms", pre=True)
    def terms_as_sequence(cls, v, values) -> Sequence[Any]:
        """Convert "terms" field to a sequence."""
        if isinstance(v, Mapping):
            v = values["predicate"].template.get_term_sequence_from_mapping(v)
        if not v:
            v = []
        elif isinstance(v, Term):
            v = [v]
        return v

    @property
    def term_sequence(self) -> TermSequence:
        """Return a TermSequence of the terms in this Statement."""
        return TermSequence(self.terms)

    @property
    def terms_without_nulls(self) -> Sequence[Term]:
        """Get Terms that are not None."""
        return [term for term in self.terms if term is not None]

    @property
    def slug(self) -> str:
        """
        Get a representation of self without whitespace.

        Intended for use as a sympy :class:`~sympy.core.symbol.Symbol`
        """
        terms = [term for term in self.terms if term is not None]
        subject = self.predicate._content_with_terms(terms).removesuffix(" was")
        return slugify(subject)

    @property
    def str_with_concrete_context(self) -> str:
        """
        Identify self more verbosely, specifying which text is a concrete context factor.

        :returns:
            the same as the __str__ method, but with an added "SPECIFIC CONTEXT" section
        """
        text = str(self)
        concrete_context = [
            factor for factor in self.terms_without_nulls if not factor.generic
        ]
        if any(concrete_context) and not self.generic:
            text += "\n" + indented("SPECIFIC CONTEXT:")
            for factor in concrete_context:
                factor_text = indented(factor.wrapped_string, tabs=2)
                text += f"\n{factor_text}"
        return text

    @property
    def truth(self) -> Optional[bool]:
        """Access :attr:`~Predicate.truth` attribute."""
        return self.predicate.truth

    @validator("terms")
    def _validate_terms(cls, v, values, **kwargs):
        """Normalize ``terms`` to initialize Statement."""

        # make TermSequence for validation, then ignore it
        TermSequence.validate_terms(v)

        if values.get("predicate") is None:
            raise ValidationError("Predicate field is required.")

        if len(v) != len(values["predicate"]):
            message = (
                "The number of items in 'terms' must be "
                + f"{len(values['predicate'])}, not {len(v)}, "
                + f"to match predicate.context_slots for '{values['predicate']}'"
            )
            raise ValueError(message)
        return v

    @property
    def wrapped_string(self):
        """Wrap text in string representation of ``self``."""
        content = str(self.predicate._content_with_terms(self.terms))
        unwrapped = self.predicate._add_truth_to_content(content)
        text = wrapped(super().__str__().format(unwrapped))
        if self.standard_of_proof:
            text += "\n" + indented(f"by the STANDARD {self.standard_of_proof}")
        return text

    @validator("standard_of_proof")
    def validate_standard_of_proof(cls, v: Optional[str]) -> Optional[str]:
        """
        Validate that standard of proof is one of the allowable values.

        :param v:
            the value to validate.

        :returns:
            the validated value.
        """
        if v is None:
            return v
        if v not in cls.standards_of_proof:
            raise ValueError(
                f"standard of proof must be one of {cls.standards_of_proof} or None."
            )
        return v

    def __str__(self):
        """Create one-line string representation for inclusion in other Facts."""
        unwrapped = self.predicate._add_truth_to_content(self.content)
        standard = (
            f"by the standard {self.standard_of_proof}, "
            if self.standard_of_proof
            else ""
        )
        string = f"{standard}{unwrapped}"
        return Comparable.__str__(self).format(string).replace(",,", ",")

    @property
    def content(self) -> str:
        """Return the content of self's Predicate."""
        return str(self.predicate._content_with_terms(self.terms))

    def _means_if_concrete(
        self, other: Comparable, context: Explanation
    ) -> Iterator[Explanation]:
        if self.standard_of_proof == other.__dict__.get(
            "standard_of_proof"
        ) and self.predicate.means(other.predicate):
            yield from super()._means_if_concrete(other, context)

    def __len__(self):
        return len(self.generic_terms())

    def _implies_if_concrete(
        self, other: Comparable, context: ContextRegister
    ) -> Iterator[ContextRegister]:
        """
        Test if ``self`` implies ``other``, assuming they are not ``generic``.

        :returns:
            whether ``self`` implies ``other`` under the given assumption.
        """
        if (
            isinstance(other, self.__class__)
            and self.predicate >= other.predicate
            and bool(self.standard_of_proof) == bool(other.standard_of_proof)
            and not (
                self.standard_of_proof
                and (
                    self.standards_of_proof.index(self.standard_of_proof)
                    < self.standards_of_proof.index(other.standard_of_proof)
                )
            )
        ):
            yield from super()._implies_if_concrete(other, context)

    def _contradicts_if_present(
        self, other: Comparable, explanation: Explanation
    ) -> Iterator[Explanation]:
        """
        Test if ``self`` contradicts :class:`Fact` ``other`` if neither is ``absent``.

        :returns:
            whether ``self`` and ``other`` can't both be true at
            the same time under the given assumption.
        """
        if isinstance(other, self.__class__) and self.predicate.contradicts(
            other.predicate
        ):
            for context in self._context_registers(
                other, operator.ge, explanation.context
            ):
                yield explanation.with_context(context)

    def negated(self) -> Fact:
        """Return copy of self with opposite truth value."""
        result = deepcopy(self)
        result.predicate = result.predicate.negated()
        return result

    @new_context_helper
    def new_context(self, changes: Dict[Comparable, Comparable]) -> Comparable:
        """
        Create new :class:`Factor`, replacing keys of ``changes`` with values.

        :returns:
            a version of ``self`` with the new context.
        """
        result = deepcopy(self)
        new_terms = TermSequence(
            [factor.new_context(changes) for factor in self.terms_without_nulls]
        )
        result.terms = list(new_terms)
        return result

    def _registers_for_interchangeable_context(
        self, matches: ContextRegister
    ) -> Iterator[ContextRegister]:
        r"""
        Find possible combination of interchangeable :attr:`terms`.

        :param matches:
            matching Terms between self and other
        :yields:
            context registers with every possible combination of
            ``self``\'s and ``other``\'s interchangeable
            :attr:`terms`.
        """
        yield matches
        gen = self.term_permutations()
        _ = next(gen)  # unchanged permutation
        already_returned: List[ContextRegister] = [matches]

        for term_permutation in gen:
            left = [term for term in self.terms if term is not None]
            right = [term for term in term_permutation if term is not None]
            changes = ContextRegister.from_lists(left, right)
            changed_registry = matches.replace_keys(changes)
            if all(
                changed_registry != returned_dict for returned_dict in already_returned
            ):
                already_returned.append(changed_registry)
                yield changed_registry

    def term_permutations(self) -> Iterator[TermSequence]:
        """Generate permutations of context factors that preserve same meaning."""
        for pattern in self.predicate.term_index_permutations():
            sorted_terms = [x for _, x in sorted(zip(pattern, self.terms))]
            yield TermSequence(sorted_terms)


def build_fact(
    predicate: Predicate,
    indices: Optional[Union[int, Sequence[int]]] = None,
    case_factors: Optional[Union[Factor, Sequence[Factor]]] = None,
    name: Optional[str] = None,
    standard_of_proof: Optional[str] = None,
    absent: bool = False,
    generic: bool = False,
):
    r"""
    Build a :class:`.Fact` with generics selected from a list.

    :param predicate:
        a natural-language clause with zero or more slots
        to insert ``terms`` that are typically the
        subject and objects of the clause.

    :param terms:
        a series of integer indices of generic factors to
        fill in the blanks in the :class:`.Predicate`

    :param name:
        an identifier for this object, often used if the object needs
        to be referred to multiple times in the process of composing
        other :class:`.Factor` objects

    :param standard_of_proof:
        a descriptor for the degree of certainty associated
        with the assertion in the :class:`.Predicate`

    :param absent:
        whether the absence, rather than the presence, of the legal
        fact described above is being asserted.

    :param generic:
        whether this object could be replaced by another generic
        object of the same class without changing the truth of the
        :class:`Rule` in which it is mentioned.

    :param case_factors:
        a series of :class:`.Factor`\s that have already been mentioned
        in the :class:`.Opinion`. They are available for composing the
        new :class:`.Factor` object and don't need to be recreated.
    """
    if not indices:
        indices = range(len(predicate))
    if isinstance(indices, int):
        indices = (indices,)

    case_factors = case_factors or ()
    if isinstance(case_factors, BaseModel):
        wrapped_factors = [case_factors]
    else:
        wrapped_factors = list(case_factors)

    terms = [wrapped_factors[i] for i in indices]
    return Fact(
        predicate=predicate,
        terms=terms,
        name=name or "",
        standard_of_proof=standard_of_proof,
        absent=absent,
        generic=generic,
    )


class Exhibit(Factor, BaseModel):
    """
    A source of information for use in litigation.

    :param form:
        a term describing the category of exhibit. For example: testimony,
        declaration, document, or recording.

    :param statement:
        a fact assertion made via the exhibit. For instance, if the exhibit
        is a document, this parameter could refer to a statement printed
        on the document.

    :param statement_attribution:
        the :class:`.Entity` that the exhibit imputes the statement to. For
        instance, for a signed declaration, this would refer to the person
        whose signature appears on the declaration, regardless of any
        authentication concerns. The statement_attribution parameter may
        appear without the statement parameter, especially if the content
        of the statement is irrelevant.

    :param name:
        a string identifier for the exhibit

    :param absent:
        if True, indicates that no exhibit meeting the description exists
        in the litigation. If the exhibit has merely been rejected as
        evidence, use the absent attribute on an :class:`Evidence` object
        instead.

    :param generic:
        if True, indicates that the specific attributes of the exhibit
        are irrelevant in the context of the :class:`.Holding` where
        the exhibit is being referenced.

    .. note
        The form parameter may be replaced by a limited
        ontology of terms when sufficient example data is available.
    """

    offered_by: Entity
    form: Optional[str] = None
    statement: Optional[Fact] = None
    statement_attribution: Optional[Entity] = None
    name: Optional[str] = None
    absent: bool = False
    generic: bool = False
    context_factor_names: ClassVar[Tuple[str, ...]] = (
        "statement",
        "statement_attribution",
    )

    def _means_if_concrete(
        self, other: Factor, context: ContextRegister
    ) -> Iterator[ContextRegister]:
        if (
            isinstance(other, self.__class__)
            and self.form == other.form
            and self.offered_by.means(other.offered_by)
        ):
            yield from super()._means_if_concrete(other, context)

    def _implies_if_concrete(
        self, other: Factor, context: Optional[ContextRegister] = None
    ) -> Iterator[ContextRegister]:
        if isinstance(other, self.__class__) and (
            self.form == other.form or other.form is None
        ):
            yield from super()._implies_if_concrete(other, context)

    def __str__(self):
        """Represent object as string without line breaks."""
        string = f'{("attributed to " + self.statement_attribution.short_string) if self.statement_attribution else ""}'
        if self.statement:
            string += ", asserting " + self.statement.short_string + ","
        string = super().__str__().format(string)
        return string.replace("exhibit", self.form or "exhibit").strip()

    @property
    def wrapped_string(self):
        """Create a string describing the Exhibit split over multiple lines."""
        text = ""
        if self.form:
            text += f"in the FORM {self.form}"
        if self.statement:
            text += "\n" + indented("WITH THE ASSERTION:")
            factor_text = indented(self.statement.wrapped_string, tabs=2)
            text += f"\n{factor_text},"
        if self.statement_attribution:
            text += "\n" + indented(
                f"ATTRIBUTED TO {self.statement_attribution.wrapped_string}"
            )

        return super().__str__().format(text)


class Evidence(Factor, BaseModel):
    """
    An :class:`Exhibit` admitted by a court to aid a factual determination.

    :param exhibit:
        the thing that is being used to aid a factual determination

    :param to_effect:
        the :class:`.Fact` finding that would be supported by the evidence.
        If the Fact object includes a non-null standard_of_proof attribute, it
        indicates that that the evidence would support a factual finding by
        that standard of proof.

    :param name:
        a string identifier

    :param absent:
        if True, indicates that no evidence meeting the description has been
        admitted, regardless of whether a corresponding :class:`Exhibit` has
        been presented

    :param generic:
        if True, indicates that the specific attributes of the evidence
        are irrelevant in the context of the :class:`.Holding` where
        the evidence is being referenced.
    """

    exhibit: Optional[Exhibit] = None
    to_effect: Optional[Fact] = None
    name: Optional[str] = None
    absent: bool = False
    generic: bool = False
    context_factor_names: ClassVar[Tuple[str, ...]] = ("exhibit", "to_effect")

    class Config:
        """Fail validation if the input data has data not specified in the model."""

        extra = Extra.forbid

    @root_validator(pre=True)
    def check_type_field(cls, values):
        """Fail valitation if the input has a "type" field without the class name."""
        type_str = values.pop("type", "")
        if type_str and type_str.lower() != "evidence":
            raise ValidationError(f"type {type_str} was passed to Evidence model")
        return values

    def __str__(self):
        string = (
            f'{("of " + self.exhibit.short_string + " ") if self.exhibit else ""}'
            + f'{("which supports " + self.to_effect.short_string) if self.to_effect else ""}'
        )
        return super().__str__().format(string).strip().replace("Evidence", "evidence")

    @property
    def wrapped_string(self):
        """Create a string describing the Evidence split over multiple lines."""
        text = ""
        if self.exhibit:
            text += "\n" + indented("OF:")
            factor_text = indented(self.exhibit.wrapped_string, tabs=2)
            text += f"\n{str(factor_text)}"
        if self.to_effect:
            text += "\n" + indented("INDICATING:")
            factor_text = indented(self.to_effect.wrapped_string, tabs=2)
            text += f"\n{factor_text}"
        return super().__str__().format(text).strip()


class Pleading(Factor, BaseModel):
    r"""
    A document filed by a party to make :class:`Allegation`\s.

    :param filer:
        the :class:`.Entity` that the pleading references as having filed it,
        regardless of any questions about the identification of the filer.

    :param name:

    :param absent:

    :param generic:
    """

    filer: Entity
    name: Optional[str] = None
    absent: bool = False
    generic: bool = False
    context_factor_names: ClassVar[Tuple[str]] = ("filer",)

    def __str__(self):
        string = f'{("filed by " + self.filer.short_string if self.filer else "")}'
        return super().__str__().format(string)


class Allegation(Factor, BaseModel):
    """
    A formal assertion of a :class:`Fact`.

    May be included by a party in a :class:`Pleading`
    to establish a cause of action.

    :param statement:
        a :class:`Fact` being alleged

    :param pleading:
        the :class:`Pleading` in where the allegation appears

    :param name:

    :param absent:

    :param generic:
    """

    fact: Fact
    pleading: Optional[Pleading] = None
    name: Optional[str] = None
    absent: bool = False
    generic: bool = False
    context_factor_names: ClassVar[Tuple[str, ...]] = ("fact", "pleading")

    @property
    def wrapped_string(self):
        """Create a string describing the Allegation split over multiple lines."""
        text = ""
        if self.fact:
            text += "\n" + indented("OF:")
            factor_text = indented(self.fact.wrapped_string, tabs=2)
            text += f"\n{str(factor_text)}"
        if self.pleading:
            text += "\n" + indented("FOUND IN:")
            factor_text = indented(str(self.pleading), tabs=2)
            text += f"\n{factor_text}"
        return super().__str__().format(text).strip()

    def __str__(self):
        string = (
            f'{("in " + self.pleading.short_string + ",") if self.pleading else ""}'
            + f'{("claiming " + self.fact.short_string + ",") if self.fact else ""}'
        )
        string = string.strip(",")
        return super().__str__().format(string).replace("Allegation", "allegation")


Fact.update_forward_refs()
Exhibit.update_forward_refs()
Evidence.update_forward_refs()
Allegation.update_forward_refs()
Pleading.update_forward_refs()
